package org.implab.gradle.mercurial;

import java.nio.charset.Charset;
import java.util.ArrayList;
import java.util.Collections;
import java.util.List;
import java.util.Objects;
import java.util.Optional;
import java.util.regex.Matcher;
import java.util.regex.Pattern;

import org.gradle.api.Project;
import org.implab.gradle.Utils;

public class MercurialExtension {
    private static final String COMMIT_INFO_TEMPLATE = "{latesttag('re:^v') % '{ifeq(tag,'null','',tag)}:{distance}'}:{branch}:{node|short}";

    private static final Pattern versionPattern = Pattern.compile("^(?:v[\\-_\\.]?)?(.*)");

    private final Project project;

    private Charset outputCharset = Charset.defaultCharset();

    private WorkspaceInfo workspaceInfo;

    private String hgCmd = "hg";

    private SemVersion defaultWorkspaceVersion = new SemVersion(0, 0, 1, "dev", null);

    public MercurialExtension(Project project) {
        Objects.requireNonNull(project);

        this.project = project;
    }

    public String hg(String... args) throws Exception {
        List<String> commandLine = new ArrayList<>();
        commandLine.add(hgCmd);
        Collections.addAll(commandLine, args);

        return Utils.execCmd(commandLine, outputCharset.name());
    }

    public Charset getOutputCharset() {
        return outputCharset;
    }

    public void setOutputCharset(Charset outputCharset) {
        this.outputCharset = Optional.ofNullable(outputCharset).orElseGet(Charset::defaultCharset);
    }

    public void setOutputCharset(String outputCharset) {
        this.outputCharset = Optional.ofNullable(outputCharset).map(x -> Charset.forName(x))
                .orElseGet(Charset::defaultCharset);
    }

    public String getHgCmd() {
        return hgCmd;
    }

    public void setHgCmd(String hgCmd) {
        this.hgCmd = hgCmd;
    }

    public WorkspaceInfo getWorkspaceInfo() throws Exception {
        if (workspaceInfo == null)
            workspaceInfo = loadWorkspaceInfo();
        return workspaceInfo;
    }

    public SemVersion getWorkspaceVersion() throws Exception {
        return tryParseVersion(getWorkspaceInfo().getVersionTag()).orElse(defaultWorkspaceVersion);
    }

    public void applyVersioningPolicy() throws Exception {
        Object version = project.getVersion();
        if (version == null || version.toString().isBlank() || version.toString().equalsIgnoreCase(Project.DEFAULT_VERSION)) {
            int distance = getWorkspaceInfo().getVersionDistance();
            String changeset = getWorkspaceInfo().getChangeset();
            project.setVersion(
                distance> 0 ?
                    getWorkspaceVersion().addPatch(distance).suffix("snapshot").meta(changeset) :
                    getWorkspaceVersion()
            );
        } else {
            project.setVersion(SemVersion.toSemVersion(version, false));
        }
    }

    String[] queryLogInfo() throws Exception {
        return hg("log", "-r", ".", "--template", COMMIT_INFO_TEMPLATE).split(":");
    }

    WorkspaceInfo loadWorkspaceInfo() throws Exception {
        boolean isDirty = checkIsDirty();
        String[] info = queryLogInfo();
        return new WorkspaceInfo(info[0], Integer.parseInt(info[1]), info[2], info[3], isDirty);
    }

    boolean checkIsDirty() throws Exception {
        return !hg("status", "-mard").isBlank();
    }

    Optional<SemVersion> tryParseVersion(String version) {
        if (version == null)
            return Optional.empty();

        Matcher m = versionPattern.matcher(version);
        boolean isMatch = m.matches();

        return isMatch ? SemVersion.tryParseLax(m.group(1)) : Optional.empty();
    }

    
}
